"""ndice - A dice rolling library for games."""

from .constants import d2, d3, d4, d6, d8, d10, d12, d20, d100
from .constants import four_d6, three_d6, two_d6
from .dice import d, Dice, mod, nd
from .mods import plus, minus, times
from .op import Op
from .rng import AscendingRNG, FixedRNG, high, low, mid, PRNG, RNG, rng
from .roll import max_roll, min_roll, roll, roll_each_die


__version__ = '0.1.9'
__author__ = 'Don McCaughey'
__email__ = 'don@donm.cc'



from .dice import Dice
from .rng import high, low, RNG


def roll(rng: RNG, *dice_expression: Dice) -> int:
    total = 0
    for dice_term in dice_expression:
        total = dice_term.op(total, sum(roll_each_die(rng, dice_term)))
    return total


def min_roll(*dice_expression: Dice) -> int:
    return roll(low, *dice_expression)


def max_roll(*dice_expression: Dice) -> int:
    return roll(high, *dice_expression)


def roll_each_die(rng: RNG, dice: Dice) -> list[int]:
    if dice.is_mod:
        return [dice.number * dice.sides]
    else:
        return [_roll_die(rng, dice.sides) for _ in range(dice.number)]


def _roll_die(rng: RNG, sides: int) -> int:
    roll = rng(sides)
    assert 1 <= roll <= sides
    return roll



from __future__ import annotations

from dataclasses import dataclass

from .interned import interned
from .op import Op


@interned
@dataclass(frozen=True, order=True, slots=True)
class Dice:
    number: int
    sides: int
    op: Op = Op.PLUS

    def __post_init__(self) -> None:
        assert self.number >= 0
        assert self.sides >= 0

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if Op.PLUS == self.op:
            return f'{cls_name}({self.number}, {self.sides})'
        else:
            return f'{cls_name}({self.number}, {self.sides}, {self.op!r})'

    def __str__(self) -> str:
        if self.is_mod:
            mod = self.number * self.sides
            return f'{self.op}{mod}'
        else:
            op = '' if Op.PLUS == self.op else self.op
            number = '' if 1 == self.number else self.number
            return f'{op}{number}d{self.sides}'

    @property
    def is_mod(self) -> bool:
        return 0 == self.number or 0 == self.sides or 1 == self.sides

    def to_plus(self) -> Dice:
        return self.__class__(self.number, self.sides, Op.PLUS)

    def to_minus(self) -> Dice:
        return self.__class__(self.number, self.sides, Op.MINUS)

    def to_times(self) -> Dice:
        return self.__class__(self.number, self.sides, Op.TIMES)

    @classmethod
    def die(cls, sides: int) -> Dice:
        return cls(1, sides)

    @classmethod
    def mod(cls, op: Op, value: int) -> Dice:
        return cls(value, 1, op)

    @classmethod
    def n_dice(cls, number: int, sides: int | Dice) -> Dice:
        sides_value = sides.sides if isinstance(sides, Dice) else sides
        return cls(number, sides_value, Op.PLUS)


d = Dice.die
mod = Dice.mod
nd = Dice.n_dice



from .dice import Dice, mod
from .op import Op


def plus(value: int | Dice) -> Dice:
    return value.to_plus() if isinstance(value, Dice) else mod(Op.PLUS, value)


def minus(value: int | Dice) -> Dice:
    return value.to_minus() if isinstance(value, Dice) else mod(Op.MINUS, value)


def times(value: int | Dice) -> Dice:
    return value.to_times() if isinstance(value, Dice) else mod(Op.TIMES, value)



from random import randrange, Random
from typing import Callable, TypeAlias


RNG: TypeAlias = Callable[[int], int]


def rng(sides: int) -> int:
    return randrange(sides) + 1


def high(sides: int) -> int:
    return sides


def low(sides: int) -> int:
    return 1


def mid(sides: int) -> int:
    return (sides + 1) // 2


def AscendingRNG(initial_value: int) -> RNG:
    ascending_value = initial_value

    def ascending(sides: int) -> int:
        nonlocal ascending_value
        next_value = (max(ascending_value, 1) - 1) % sides + 1
        ascending_value += 1
        return next_value

    return ascending


def FixedRNG(fixed_value: int) -> RNG:
    return lambda sides: _clamp(1, fixed_value, sides)


def PRNG(seed: int) -> RNG:
    r = Random(seed)
    return lambda sides: r.randrange(sides) + 1


def _clamp(lower: int, value: int, upper: int) -> int:
    return max(lower, min(value, upper))
